import os
from typing import Any, Callable, Optional
import cv2
import numpy as np


class VideoSaver:
    def __init__(self, output_name: str, fps: int = 30):
        self.writer: cv2.writer
        self.empty: bool = True
        self.fps = fps
        self.output_path = output_name

    def new_writer(self, w: int, h: int):
        self.writer = cv2.VideoWriter(
            self.output_path, cv2.VideoWriter_fourcc(*"MP4V"), self.fps, (w, h)
        )
        self.empty = False

    def add_frame(self, frame: cv2.Mat):
        if self.empty:
            self.new_writer(*frame.shape[:2][::-1])
        self.writer.write(frame.astype(np.uint8))

    def finish(self):
        self.writer.release()


class PipelineVideoSaver:
    def __init__(self, img_to_path_function: Callable[[Any], str]):
        self.vs: Optional[VideoSaver] = None
        self.img_to_path_function = img_to_path_function

    def _ensure_folder(self, path: str):
        folder = os.path.dirname(path)
        os.makedirs(folder, exist_ok=True)

    def process_file(self, file: Any):
        path = self.img_to_path_function(file)
        self._ensure_folder(path)
        if self.vs is None:
            self.vs = VideoSaver(path)
            self.vs.add_frame(file.img)
        elif self.vs.output_path == path:
            self.vs.add_frame(file.img)
        else:
            self.vs.finish()
            self.vs = VideoSaver(path)
            self.vs.add_frame(file.img)

    def finish(self):
        if self.vs is not None:
            self.vs.finish()



from typing import List, Tuple
import cv2
import numpy as np

Size = Tuple[int, int]
Color = Tuple[int, int, int]
Image = cv2.Mat


class ImTools:
    @staticmethod
    def load(path: str) -> Image:
        return cv2.imread(path)

    @staticmethod
    def load_multiple(paths: List[str]) -> List[Image]:
        return [ImTools.load(path) for path in paths]

    @staticmethod
    def pad_square(image: Image, color: Color = (0, 0, 0)) -> Image:
        """
        Pads an image to fit in a square
        Args:
            image (Image): cv2.Image
        Returns:
            Image: cv2.Image
        """
        max_dim = max(image.shape[:-1])
        padded_image = ImTools.pad_to_output_size(image, (max_dim, max_dim), color)
        return padded_image

    @staticmethod
    def pad_to_ratio(image: Image, ratio: float, color: Color = (0, 0, 0)) -> Image:
        """
        Pads an image to the specified ratio

        Args:
            image (Image): cv2.Image
            ratio (float): width / height

        Returns:
            Image: cv2.Image
        """
        height, width = image.shape[:-1]
        image_ratio = width / height
        if image_ratio < ratio:
            # keep height
            dims = (height, int(round(height * ratio)))
        else:
            # keep width
            dims = (int(round(width / ratio)), width)

        return ImTools.pad_to_output_size(image, dims, color)

    @staticmethod
    def pad_to_output_size(
        image: Image,
        size: Size,
        color: Color = (0, 0, 0),
        interpolation: int = cv2.INTER_LINEAR,
    ) -> Image:
        """Pads an image to the specified size (adds black pixels)

        Args:
            image (Image): cv2.Image
            size (Size): height, width

        Returns:
            Image: cv2.Image
        """
        height, width = image.shape[:-1]

        out_h, out_w = size

        desired_ratio = out_w / out_h
        input_ratio = width / height

        if input_ratio > desired_ratio:
            resized = cv2.resize(
                image, (out_w, int(out_w / input_ratio)), interpolation=interpolation
            )
        else:
            resized = cv2.resize(
                image, (int(out_h * input_ratio), out_h), interpolation=interpolation
            )

        resized_h, resized_w, _ = resized.shape

        pad_top = int((out_h - resized_h) / 2)
        pad_left = int((out_w - resized_w) / 2)
        pad_bottom = int(out_h - pad_top - resized_h)
        pad_right = int(out_w - pad_left - resized_w)

        padded_image = cv2.copyMakeBorder(
            resized,
            pad_top,
            pad_bottom,
            pad_left,
            pad_right,
            cv2.BORDER_CONSTANT,
            value=color,
        )

        return padded_image

    @staticmethod
    def crop_image(
        image: Image,
        left: int = 0,
        top: int = 0,
        right: int = 0,
        bottom: int = 0,
        pad_color: Color = (0, 0, 0),
    ) -> Image:
        """Crops an image. It can also pad an image if numbers are < 0

        Args:
            image (Image): Input cv2 image
            top (int, optional): pixels to remvoe from the top. Defaults to 0.
            left (int, optional): pixels to remvoe from the left. Defaults to 0.
            bottom (int, optional): pixels to remvoe from the bottom. Defaults to 0.
            right (int, optional): pixels to remvoe from the right. Defaults to 0.
            pad_color (Color, optional): Color to use if padding.

        Returns:
            Image: The cv2 image cropped
        """

        h, w, channels = image.shape

        new_h = h - top - bottom
        new_w = w - left - right

        img = np.ones((new_h, new_w, channels), dtype=np.uint8) * pad_color

        left_source = max(left, 0)
        left_destination = max(-left, 0)

        top_source = max(top, 0)
        top_destination = max(-top, 0)

        copy_w = w - max(left, 0) - max(right, 0)
        copy_h = h - max(top, 0) - max(bottom, 0)

        img[
            top_destination : top_destination + copy_h,
            left_destination : left_destination + copy_w,
            :,
        ] = image[
            top_source : top_source + copy_h,
            left_source : left_source + copy_w,
            :,
        ]

        return img

    @staticmethod
    def draw_heatmap(image: Image, map: Image, alpha_map: float = 0.5) -> Image:
        """
        Args:
            image (Image): cv2.image (HWC)
            map (Image): [0-255] cv2.image (HW)
            alpha_map (float): [0-1] balance between image and map
        """

        heatmap_img = cv2.applyColorMap(np.array(map, dtype=np.uint8), cv2.COLORMAP_JET)

        heatmap_img = cv2.resize(
            heatmap_img, image.shape[:2][::-1], interpolation=cv2.INTER_LINEAR
        )

        output = cv2.addWeighted(
            heatmap_img,
            alpha_map,
            image,
            1 - alpha_map,
            0,
            dtype=cv2.PARAM_UNSIGNED_INT,
        )

        return output

    @staticmethod
    def draw_multiple_heatmaps(
        image: Image, maps: Image, alpha_map: float = 0.5
    ) -> Image:
        """
        Args:
            image (Image): cv2.image (HWC)
            map (Image): [0-255] cv2.image (HW)
            alpha_map (float): [0-1] balance between image and map
        """
        max_map: Image = np.max(maps, axis=-1)
        return ImTools.draw_heatmap(image, max_map, alpha_map=alpha_map)



import os
import shutil
from glob import glob
from typing import (
    Any,
    Callable,
    Dict,
    Generator,
    Iterable,
    List,
    Literal,
    Optional,
    Tuple,
    cast,
)

import cv2
import numpy as np
from olympipe import Pipeline

from olympict.video_saver import PipelineVideoSaver

from .image_tools import ImTools

# from nptyping import NDArray, Int, Shape
# Img = NDArray[Shape["2, 2"], Int]

Img = cv2.Mat
Point = Tuple[float, float]
Size = Tuple[int, int]
Color = Tuple[int, int, int]
ImgFormat = Literal["png", "jpeg", "bmp"]
VidFormat = Literal["mp4", "mkv", "avi"]


class OlympFile:
    def __init__(self, path: Optional[str] = None):
        self.path = path or ""

    @property
    def size(self) -> Size:
        raise NotImplementedError("No size for this object")


class OlympImage(OlympFile):
    __id = 0

    def __init__(self, path: Optional[str] = None):
        super().__init__(path)
        if path is None:
            self.path = f"./{self.__id}.png"
            self.__id += 1
            self.img: Img = np.zeros((1, 1, 3), dtype=np.uint8)
        else:
            self.img: Img = cv2.imread(path)
        self.metadata: Dict[str, Any] = {}

    @staticmethod
    def resize(
        size: Size,
        pad_color: Optional[Color] = None,
        interpolation: int = cv2.INTER_LINEAR,
    ) -> Callable[["OlympImage"], "OlympImage"]:
        def r(o: "OlympImage") -> "OlympImage":
            if pad_color is None:
                o.img = cv2.resize(o.img, size, interpolation=interpolation)
            else:
                o.img = ImTools.pad_to_output_size(
                    o.img, size, pad_color, interpolation=interpolation
                )
            return o

        return r

    @staticmethod
    def rescale(
        scales: Tuple[float, float],
        pad_color: Optional[Color] = None,
        interpolation: int = cv2.INTER_LINEAR,
    ) -> Callable[["OlympImage"], "OlympImage"]:
        """Rescale function
        This function applies a scale [x_s, y_s] to a given image
        The resulting image dimensions are [w * x_s, h * y_s]
        """

        def r(o: "OlympImage") -> "OlympImage":
            w, h = o.size
            x_scale, y_scale = scales
            size = (int(round(w * x_scale)), int(round(h * y_scale)))
            if pad_color is None:
                o.img = cv2.resize(o.img, size, interpolation=interpolation)
            else:
                o.img = ImTools.pad_to_output_size(
                    o.img, size, pad_color, interpolation=interpolation
                )
            return o

        return r

    def move_to_path(self, path: str):
        """This function moves images to a new location. If path is a directory, then it will keep its old name and move to the new directory.
        Else it will be given path as a new name (This might be bad for multiple images).
        """
        #  TODO: Ensure folder or not
        if os.path.isdir(path):
            _, filename = os.path.split(self.path)
            path = os.path.join(path, filename)
        shutil.move(self.path, path)
        self.path = os.path.abspath(path)

    def change_folder_path(self, new_folder_path: str):
        self.path = os.path.join(new_folder_path, os.path.basename(self.path))

    def move_to(self, func: Callable[[str], str]):
        output = func(self.path)
        self.move_to_path(output)

    def save(self):
        os.makedirs(os.path.dirname(self.path), exist_ok=True)
        _ = cv2.imwrite(self.path, self.img)

    def save_as(self, path: str):
        if os.path.isdir(path):
            _, filename = os.path.split(self.path)
            path = os.path.join(path, filename)

        self.path = os.path.abspath(path)
        self.save()

    @property
    def size(self) -> Size:
        h, w, _ = self.img.shape
        return (w, h)

    @staticmethod
    def load(path: str, metadata: Dict[str, Any] = {}) -> "OlympImage":
        o = OlympImage()
        o.path = path
        o.img = cv2.imread(o.path)
        o.metadata = metadata
        return o

    @staticmethod
    def from_buffer(
        buffer: Img, path: str = "", metadata: Dict[str, Any] = {}
    ) -> "OlympImage":
        o = OlympImage()
        o.path = path
        o.img = buffer
        o.metadata = metadata
        return o


class OlympBatch(OlympFile):
    def __init__(self, data: Img, metadata: List[Dict[str, Any]] = []):
        assert len(data.shape) >= 4
        self.data = data
        self.metadata = metadata

    @property
    def size(self) -> Size:
        _, h, w, _ = self.data.shape
        return (w, h)

    @staticmethod
    def from_images(images: List[OlympImage]) -> "OlympBatch":
        data = np.array([i.img for i in images])
        metadata = [i.metadata for i in images]

        return OlympBatch(data, metadata)


class OlympVid(OlympFile):
    __id = 0

    def __init__(self, path: Optional[str] = None):
        super().__init__(path)
        if path is None:
            self._tmp_path = f"./{self.__id}.mp4"
            self.__id += 1
        self._tmp_path = ""
        self._fps = 25

    def get_fps(
        self,
    ) -> int:
        raise NotImplementedError()

    def to_format(self, fmt: VidFormat) -> "OlympImage":
        raise NotImplementedError()

    def get_temp_path(self) -> str:
        raise NotImplementedError()


class ObjPipeline:
    def resize(self, size: Size, count: int = 1):
        """
        Args:
            size (Size): width, height
        """

    def from_generator(self, generator: Generator[Img, None, None]) -> "ObjPipeline":
        raise NotImplementedError()

    def rescale(self, size: Tuple[float, float], count: int = 1) -> "ObjPipeline":
        """
        Args:
            size (Tuple[float, float]): width multiplier, height multiplier
        """
        raise NotImplementedError()

    def crop(self, size: Size, count: int = 1) -> "ObjPipeline":
        """
        Args:
            size (Size): width, height
        """
        raise NotImplementedError()

    def img_task(self, img_func: Callable[[Img], Img]) -> "ObjPipeline":
        raise NotImplementedError()

    def path_task(self, path_func: Callable[[str], str]) -> "ObjPipeline":
        raise NotImplementedError()


class OPipeline:
    def __init__(self, source: Iterable[OlympFile]):
        self._data = source
        self._pipeline: Optional[Pipeline[OlympFile]] = None

    @staticmethod
    def load_folder(
        path: str, extensions: List[str] = [""], recursive: bool = False
    ) -> "OPipeline":
        raise NotImplementedError()

    def wait_for_completion(self) -> None:
        if self._pipeline is None:
            return
        return self._pipeline.wait_for_completion()

    def wait_for_results(self) -> List[OlympFile]:
        if self._pipeline is None:
            return []
        return self._pipeline.wait_for_results()


class ImagePipeline(OPipeline):
    def __init__(self, data: Optional[Iterable[OlympImage]] = None):
        self._pipeline: Optional[Pipeline[OlympImage]] = None
        if data is not None:
            self._pipeline = Pipeline(data)

    @staticmethod
    def load_folder(
        path: str,
        extensions: List[str] = ["png", "jpg", "jpeg", "bmp"],
        recursive: bool = False,
        order_func: Optional[Callable[[str], int]] = None,
        reverse: bool = False,
        metadata_function: Optional[Callable[[str], Any]] = None,
    ) -> "ImagePipeline":
        paths: List[str] = glob(os.path.join(path, "**"), recursive=recursive)

        paths = [p for p in paths if os.path.splitext(p)[1].strip(".") in extensions]

        if order_func is not None:
            paths.sort(key=order_func, reverse=reverse)

        data = [OlympImage(p) for p in paths]

        if metadata_function is not None:
            for d in data:
                d.metadata = metadata_function(d.path)

        return ImagePipeline(data)

    @staticmethod
    def load_folders(
        paths: List[str],
        extensions: List[str] = ["png", "jpg", "jpeg", "bmp"],
        recursive: bool = False,
        order_func: Optional[Callable[[str], int]] = None,
        reverse: bool = False,
        metadata_function: Optional[Callable[[str], Any]] = None,
    ) -> "ImagePipeline":
        all_data: List[OlympImage] = []
        for path in paths:
            sub_paths: List[str] = glob(os.path.join(path, "**"), recursive=recursive)

            sub_paths = [
                p for p in sub_paths if os.path.splitext(p)[1].strip(".") in extensions
            ]

            if order_func is not None:
                sub_paths.sort(key=order_func, reverse=reverse)

            data = [OlympImage(p) for p in sub_paths]

            if metadata_function is not None:
                for d in data:
                    d.metadata = metadata_function(d.path)

            all_data.extend(data)

        return ImagePipeline(all_data)

    def task(
        self, func: Callable[[OlympImage], OlympImage], count: int = 1
    ) -> "ImagePipeline":
        output = ImagePipeline()
        if self._pipeline is None:
            raise Exception("Undefined pipeline")
        output._pipeline = self._pipeline.task(func, count)

        return output

    def task_img(self, func: Callable[[Img], Img], count: int = 1) -> "ImagePipeline":
        def r(o: OlympImage) -> OlympImage:
            o.img = func(o.img)
            return o

        return self.task(r, count)

    def task_path(self, func: Callable[[str], str], count: int = 1) -> "ImagePipeline":
        def r(o: OlympImage) -> OlympImage:
            o.path = func(o.path)
            return o

        return self.task(r, count)

    def rescale(
        self,
        size: Tuple[float, float],
        pad_color: Optional[Tuple[int, int, int]] = None,
        count: int = 1,
    ) -> "ImagePipeline":
        return self.task(OlympImage.rescale(size, pad_color), count)

    def resize(
        self,
        size: Tuple[int, int],
        pad_color: Optional[Tuple[int, int, int]] = None,
        interpolation: int = cv2.INTER_LINEAR,
        count: int = 1,
    ) -> "ImagePipeline":
        return self.task(OlympImage.resize(size, pad_color, interpolation), count)

    def crop(
        self,
        left: int = 0,
        top: int = 0,
        right: int = 0,
        bottom: int = 0,
        pad_color: Color = (0, 0, 0),
    ) -> "ImagePipeline":
        def r(img: Img) -> Img:
            return ImTools.crop_image(
                img, top=top, left=left, bottom=bottom, right=right, pad_color=pad_color
            )

        return self.task_img(r)

    def keep_each_frame_in(
        self, keep_n: int = 1, discard_n: int = 0
    ) -> "ImagePipeline":
        if self._pipeline is None:
            raise Exception("No defined pipeline")

        def discarder():
            while True:
                for _ in range(keep_n):
                    yield True
                for _ in range(discard_n):
                    yield False

        d = discarder()

        def get_next(_: Any) -> bool:
            return next(d)

        output = ImagePipeline()
        output._pipeline = self._pipeline.filter(get_next)

        return output

    def debug_window(self) -> "ImagePipeline":
        def d(o: "OlympImage") -> "OlympImage":
            return o

        return self.task(d)

    def to_video(
        self, img_to_video_path: Callable[[OlympImage], str]
    ) -> "VideoPipeline":
        output = VideoPipeline()
        output._pipeline = self._pipeline.class_task(
            PipelineVideoSaver,
            PipelineVideoSaver.process_file,
            [img_to_video_path],
            PipelineVideoSaver.finish,
        )

        return output

    def to_format(self, format: ImgFormat) -> "ImagePipeline":
        def change_format(path: str) -> str:
            base, _ = os.path.splitext(path)

            fmt = f".{format}" if "." != format[0] else format

            return base + fmt

        return self.task_path(change_format)

    def save_to_folder(self, folder_path: str) -> "ImagePipeline":
        os.makedirs(folder_path, exist_ok=True)

        def s(o: "OlympImage") -> "OlympImage":
            o.change_folder_path(folder_path)
            o.save()
            return o

        return self.task(s)

    def save(self) -> "ImagePipeline":
        def s(o: "OlympImage") -> "OlympImage":
            o.save()
            return o

        return self.task(s)


class VideoPipeline(OPipeline):
    def __init__(self, data: Optional[Iterable[OlympVid]] = None):
        self._pipeline: Optional[Pipeline[OlympVid]] = None
        if data is not None:
            self._pipeline = Pipeline(data)

    @staticmethod
    def load_folder(
        path: str,
        extensions: List[str] = ["mkv", "mp4"],
        recursive: bool = False,
        order_func: Optional[Callable[[str], int]] = None,
        reverse: bool = False,
    ) -> "VideoPipeline":
        paths: List[str] = glob(os.path.join(path, "**"), recursive=recursive)
        paths = [p for p in paths if os.path.splitext(p)[1].strip(".") in extensions]

        if order_func is not None:
            paths.sort(key=order_func, reverse=reverse)

        data = [OlympVid(p) for p in paths]

        return VideoPipeline(data)

    def task(
        self, func: Callable[[OlympVid], OlympVid], count: int = 1
    ) -> "VideoPipeline":
        raise NotImplementedError()

    def split_duration(self, duration: float) -> "VideoPipeline":
        raise NotImplementedError()

    def change_fps(self, fps: int) -> "VideoPipeline":
        raise NotImplementedError()

    def move_to_folder(self, path: str) -> "VideoPipeline":
        raise NotImplementedError()

    def move_to(self, path: str) -> "VideoPipeline":
        raise NotImplementedError()

    def format(self, format: Literal["mkv"]) -> "VideoPipeline":
        raise NotImplementedError()

    def to_sequence(self) -> "ImagePipeline":
        if self._pipeline is None:
            raise Exception("No defined pipeline")

        def generator(o: "OlympVid") -> Generator[OlympImage, None, None]:
            capture: Any = cv2.VideoCapture(o.path)
            res, frame = cast(Tuple[bool, Img], capture.read())
            idx = 0
            while res:
                new_path = f"{o.path}_{idx}.png"
                yield OlympImage.from_buffer(
                    frame, new_path, {"video_path": o.path, "video_frame": idx}
                )
                res, frame = cast(Tuple[bool, Img], capture.read())
                idx += 1

        output = ImagePipeline()
        output._pipeline = self._pipeline.explode(generator)

        return output

    def draw_polygons(self, polygons: List[List[Point]]) -> "ImagePipeline":
        raise NotImplementedError()

    def draw_polygon(self, polygon: List[Point]) -> "ImagePipeline":
        raise NotImplementedError()

    def draw_heatmap(self, heatmap_func: Callable[[Img], Img]) -> "ImagePipeline":
        raise NotImplementedError()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['olympict']

package_data = \
{'': ['*']}

install_requires = \
['olympipe>=1.0.3', 'opencv-python>=4.6.0.66,<5.0.0.0']

setup_kwargs = {
    'name': 'olympict',
    'version': '0.1.0',
    'description': 'A powerful parallel pipelining tool for image processing',
    'long_description': '# Olympict\n\n![coverage](https://gitlab.com/gabraken/olympict/badges/master/coverage.svg?job=tests)![status](https://gitlab.com/gabraken/olympict/badges/master/pipeline.svg)\n\n![Olympict](https://gitlab.com/gabraken/olympict/-/raw/master/Olympict.png)\n\n\nThis project will make image processing pipelines \neasy to use using the basic multiprocessing module. This module uses type checking to ensure your data process validity from the start.\n\n',
    'author': 'GKasser',
    'author_email': 'gabriel.kasser@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://gitlab.com/gabraken/olympict',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
